﻿#region -- Copyright --
//-----------------------------------------------------------------------------
// Copyright (C) 2009 Informatica Ambientale Srl. (http://www.iambientale.it)
// All rights reserved.
//
// This source code is intended for internal use only as part of the
// "Voronoi Delaunay EMGU Library Test" Development Project and/or Documentation.
//
// THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY OF ANY KIND,
// EITHER EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
// WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT AND/OR FITNESS
// FOR A PARTICULAR PURPOSE.
//-----------------------------------------------------------------------------

//----------------------------------------------------------------
// Author: Ariel Serrano
// Created: 10/28/2009
//
// Last Modified: 11/01/2009 (Ariel Serrano)
//----------------------------------------------------------------
#endregion 

using System;
using System.Collections.Generic;
using System.Drawing;
using System.IO;
using System.Windows.Forms;
using Emgu.CV;
using Emgu.CV.Structure;
using ThinkGeo.MapSuite.Core;

namespace Voronoi_EMGU_ConsoleApplication
{
    /// <summary>
    /// 
    /// </summary>
    class Program
    {
        /// <summary>
        /// 
        /// </summary>
        const float MAX_VALUE = 800;
        const float NUM_POINTS = 500;

        /// <summary>
        /// Mains the specified args.
        /// </summary>
        /// <param name="args">The args.</param>
        static void Main(string[] args)
        {
            Application.EnableVisualStyles();
            Application.SetCompatibleTextRenderingDefault(false);
            DoVoronoiDelaunay();
        }

        /// <summary>
        /// Does the voronoi diagram and delaunay triangulation.
        /// </summary>
        static void DoVoronoiDelaunay()
        {
            string shapeFileName = "InputData.shp";
            string voronoiShapeFileName = "voronoi.shp";
            string delaunayShapeFileName = "delaunay.shp";

            // create the point type shape file to be used as input data
            CreatePointShape(shapeFileName);

            // get the input PointF values from PointShape features on shape
            PointF[] points = GetPointsFromShape(shapeFileName);

            // create the bounding box shape, in this case a rectangle shape is enough
            RectangleShape boundingShape = new RectangleShape(-10, MAX_VALUE + 10, MAX_VALUE + 10, -10);

            // create a new shape with the specified name and cut it to fit inside the bounding shape if specified
            CreateVoronoiDiagramShape(voronoiShapeFileName, points, boundingShape);

            // create a new shape with the specified name and cut it to fit inside the bounding shape if specified
            CreateDelaunayTriangulationShape(delaunayShapeFileName, points, boundingShape);

            // stop the console to see the statistics
            Console.ReadKey();
        }

        /// <summary>
        /// Creates the delaunay triangulation shape.
        /// </summary>
        /// <param name="delaunayShapeFileName">Name of the delaunay shape file.</param>
        /// <param name="points">The points.</param>
        /// <param name="boundingShape">The bounding shape.</param>
        private static void CreateDelaunayTriangulationShape(string delaunayShapeFileName, PointF[] points, RectangleShape boundingShape)
        {
            if (points != null && points.Length > 0)
            {
                DateTime startTime, endTime;
                startTime = DateTime.Now;
                Console.WriteLine("Creating Delaunay Triangulation ...");

                Triangle2DF[] delaunayTriangles;
                using (PlanarSubdivision subdivision = new PlanarSubdivision(points))
                {
                    //Obtain the delaunay's triangulation from the set of points;
                    delaunayTriangles = subdivision.GetDelaunayTriangles();
                }

                endTime = DateTime.Now;
                Console.WriteLine("Delaunay Triangulation creation process finished");
                TimeSpan duration = endTime - startTime;
                Console.WriteLine("Proccess duration: " + duration.ToString());
                Console.WriteLine("");

                // Create delaunay's triangulation shape file and cut it to fit inside the bounding shape if specified
                CreateDelaunayShapeFromTriangulation(delaunayTriangles, delaunayShapeFileName, boundingShape);
            }
        }

        /// <summary>
        /// Creates the delaunay shape from triangulation.
        /// </summary>
        /// <param name="delaunayTriangles">The delaunay triangles.</param>
        /// <param name="delaunayShapeFileName">Name of the delaunay shape file.</param>
        /// <param name="boundingShape">The bounding shape.</param>
        private static void CreateDelaunayShapeFromTriangulation(Triangle2DF[] delaunayTriangles, string delaunayShapeFileName, RectangleShape boundingShape)
        {
            Console.WriteLine("Creating delaunay triangulation shape file " + delaunayShapeFileName + " ...");

            // define the dbf columns
            DbfColumn LATColumn = new DbfColumn("LAT", DbfColumnType.Double, 30, 10);
            DbfColumn LONGColumn = new DbfColumn("LONG", DbfColumnType.Double, 30, 10);

            // create the shape file
            ShapeFileFeatureSource.CreateShapeFile(
                ShapeFileType.Polygon,
                delaunayShapeFileName,
                new DbfColumn[] { LATColumn, LONGColumn },
                System.Text.Encoding.UTF8,
                OverwriteMode.Overwrite);


            // open the shape file in read / write mode
            ShapeFileFeatureSource vShapeFile = new ShapeFileFeatureSource(delaunayShapeFileName, ShapeFileReadWriteMode.ReadWrite);

            // define the column dictionary that will hold the columns values
            Dictionary<string, string> columnsDic = new Dictionary<string, string>();
            columnsDic.Add("LAT", "");
            columnsDic.Add("LONG", "");

            try
            {
                vShapeFile.Open();
                vShapeFile.BeginTransaction();

                //Draw the delaunay triangles
                Vertex[] vertices = new Vertex[3];
                PolygonShape polygon;
                MultipolygonShape multiPolygon;

                foreach (Triangle2DF triangle in delaunayTriangles)
                {
                    columnsDic["LAT"] = triangle.Centeroid.X.ToString();
                    columnsDic["LONG"] = triangle.Centeroid.Y.ToString();

                    vertices[0] = new Vertex(triangle.V0.X, triangle.V0.Y);
                    vertices[1] = new Vertex(triangle.V1.X, triangle.V1.Y);
                    vertices[2] = new Vertex(triangle.V2.X, triangle.V2.Y);

                    //Draw the triangle as polygon
                    polygon = new PolygonShape(new RingShape(vertices));

                    // if specified a boundary shape, cut the diagram to fit inside
                    if (boundingShape != null)
                    {
                        multiPolygon = boundingShape.GetIntersection(polygon);

                        if (multiPolygon != null && multiPolygon.Polygons.Count > 0)
                        {
                            polygon = multiPolygon.Polygons[0];
                        }
                    }

                    vShapeFile.AddFeature(new Feature(polygon, columnsDic));
                }

                vShapeFile.CommitTransaction();
            }
            finally
            {
                // close the shape file handler
                if (vShapeFile.IsOpen)
                {
                    vShapeFile.Close();
                }
            }

            Console.WriteLine("Shape file created.");
            Console.WriteLine("");
        }

        /// <summary>
        /// Creates the voronoi diagram shape with the specified name and cut it to fit inside the bounding shape if specified
        /// </summary>
        /// <param name="voronoiShapeFileName">Name of the voronoi shape file.</param>
        /// <param name="points">The points.</param>
        /// <param name="boundingShape">The bounding shape.</param>
        private static void CreateVoronoiDiagramShape(string voronoiShapeFileName, PointF[] points, AreaBaseShape boundingShape)
        {
            if (points != null && points.Length > 0)
            {
                DateTime startTime, endTime;
                startTime = DateTime.Now;
                Console.WriteLine("Creating Voronoi Diagram ...");

                VoronoiFacet[] voronoiFacets;
                using (PlanarSubdivision subdivision = new PlanarSubdivision(points))
                {
                    //Obtain the voronoi facets from the set of points
                    voronoiFacets = subdivision.GetVoronoiFacets();
                }

                endTime = DateTime.Now;
                Console.WriteLine("Voronoi Diagram creation process finished");
                TimeSpan duration = endTime - startTime;
                Console.WriteLine("Proccess duration: " + duration.ToString());
                Console.WriteLine("");

                // Create voronoi diagram shape file and cut it to fit inside the bounding shape if specified
                CreateVoronoiShapeFromDiagram(voronoiFacets, voronoiShapeFileName, boundingShape);
            }
        }

        /// <summary>
        /// Creates the voronoi shape from diagram.
        /// </summary>
        /// <param name="voronoiFacets">The voronoi facets.</param>
        /// <param name="voronoiShapeFileName">Name of the voronoi shape file.</param>
        /// <param name="boundingShape">The bounding shape.</param>
        private static void CreateVoronoiShapeFromDiagram(VoronoiFacet[] voronoiFacets, string voronoiShapeFileName, AreaBaseShape boundingShape)
        {
            Console.WriteLine("Creating voronoi shape file " + voronoiShapeFileName + " ...");

            // define the dbf columns
            DbfColumn LATColumn = new DbfColumn("LAT", DbfColumnType.Double, 30, 10);
            DbfColumn LONGColumn = new DbfColumn("LONG", DbfColumnType.Double, 30, 10);

            // create the shape file
            ShapeFileFeatureSource.CreateShapeFile(
                ShapeFileType.Polygon,
                voronoiShapeFileName,
                new DbfColumn[] { LATColumn, LONGColumn },
                System.Text.Encoding.UTF8,
                OverwriteMode.Overwrite);

            // open the shape file in read / write mode
            ShapeFileFeatureSource vShapeFile = new ShapeFileFeatureSource(voronoiShapeFileName, ShapeFileReadWriteMode.ReadWrite);

            // define the column dictionary that will hold the columns values
            Dictionary<string, string> columnsDic = new Dictionary<string, string>();
            columnsDic.Add("LAT", "");
            columnsDic.Add("LONG", "");

            try
            {
                vShapeFile.Open();
                vShapeFile.BeginTransaction();

                //Draw the voronoi Facets
                Vertex[] vertices;
                PolygonShape polygon;
                MultipolygonShape multiPolygon;

                foreach (VoronoiFacet facet in voronoiFacets)
                {
                    columnsDic["LAT"] = facet.Point.X.ToString();
                    columnsDic["LONG"] = facet.Point.Y.ToString();

                    vertices = Array.ConvertAll<PointF, Vertex>(facet.Vertices, new Converter<PointF, Vertex>((input) => { return new Vertex(input.X, input.Y); }));

                    //Draw the facet as polygon
                    polygon = new PolygonShape(new RingShape(vertices));

                    // if specified a boundary shape, cut the diagram to fit inside
                    if (boundingShape != null)
                    {
                        multiPolygon = boundingShape.GetIntersection(polygon);

                        if (multiPolygon != null && multiPolygon.Polygons.Count > 0)
                        {
                            polygon = multiPolygon.Polygons[0];
                        }
                    }

                    vShapeFile.AddFeature(new Feature(polygon, columnsDic));
                }

                vShapeFile.CommitTransaction();
            }
            finally
            {
                // close the shape file handler
                if (vShapeFile.IsOpen)
                {
                    vShapeFile.Close();
                }
            }

            Console.WriteLine("Shape file created.");
            Console.WriteLine("");
        }

        /// <summary>
        /// Gets the points from shape.
        /// </summary>
        /// <param name="shapeFileName">Name of the shape file.</param>
        /// <returns></returns>
        private static PointF[] GetPointsFromShape(string shapeFileName)
        {
            List<PointF> points = new List<PointF>();

            if (File.Exists(shapeFileName))
            {
                ShapeFileFeatureLayer inputShapeFile = null;

                try
                {
                    // open the shape file in read mode
                    inputShapeFile = new ShapeFileFeatureLayer(shapeFileName, ShapeFileReadWriteMode.ReadOnly);
                    inputShapeFile.Open();

                    PointShape pointShape;
                    foreach (Feature feature in inputShapeFile.FeatureSource.GetAllFeatures(ReturningColumnsType.NoColumns))
                    {
                        pointShape = feature.GetShape() as PointShape;

                        if (pointShape != null)
                        {
                            points.Add(new PointF((float)pointShape.X, (float)pointShape.Y));
                        }
                    }
                }
                finally
                {
                    // close the shape file handler
                    if (inputShapeFile != null && inputShapeFile.IsOpen)
                    {
                        inputShapeFile.Close();
                    }
                }
            }

            return points.ToArray();
        }

        /// <summary>
        /// Creates the point shape.
        /// </summary>
        /// <returns>the shape file name</returns>
        private static void CreatePointShape(string shapeFileName)
        {
            // define the Dbf columns
            DbfColumn LATColumn = new DbfColumn("LAT", DbfColumnType.Double, 30, 10);
            DbfColumn LONGColumn = new DbfColumn("LONG", DbfColumnType.Double, 30, 10);

            Console.WriteLine("Creating shape file " + shapeFileName + " ...");

            // create the shape file
            ShapeFileFeatureSource.CreateShapeFile(
                ShapeFileType.Point,
                shapeFileName,
                new DbfColumn[] {LATColumn, LONGColumn},
                System.Text.Encoding.UTF8,
                OverwriteMode.Overwrite);

            // open the shape file in read / write mode
            ShapeFileFeatureSource inputShapeFile = new ShapeFileFeatureSource(shapeFileName, ShapeFileReadWriteMode.ReadWrite);

            // define the column dictionary that will hold the columns values
            Dictionary<string, string> columnsDic = new Dictionary<string, string>();
            columnsDic.Add("LAT", "");
            columnsDic.Add("LONG", "");

            try
            {
                inputShapeFile.Open();
                inputShapeFile.BeginTransaction();

                double lat, lon;
                // create random points in the range of [0, MAX_VALUE] and add it to the shape file
                Random r = new Random((int)(DateTime.Now.Ticks & 0x0000ffff));
                for (int i = 0; i < NUM_POINTS; i++)
                {
                    lon = r.NextDouble() * MAX_VALUE;
                    lat = r.NextDouble() * MAX_VALUE;

                    columnsDic["LAT"] = lat.ToString();
                    columnsDic["LONG"] = lon.ToString();

                    inputShapeFile.AddFeature(new Feature(new PointShape(lon, lat), columnsDic));
                }
                
                inputShapeFile.CommitTransaction();
            }
            finally
            {
                // close the shape file handler
                if (inputShapeFile.IsOpen)
                {
                    inputShapeFile.Close();
                }
            }

            Console.WriteLine("Shape file created.");
            Console.WriteLine("");
        }
    }
}
